#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pandas as pd

from kxy.api import upload_data
from kxy.pre_learning import data_valuation as dv
from kxy.pre_learning import variable_selection as vs

from .base_accessor import BaseAccessor

@pd.api.extensions.register_dataframe_accessor("kxy_pre_learning")
class PreLearningAccessor(BaseAccessor):
	"""
	Extension of the pandas.DataFrame class with various analytics for **post-learning** in supervised learning problems.

	This class defines the :code:`kxy_pre_learning` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_.

	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_pre_learning.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
	"""
	def data_valuation(self, target_column, problem_type=None, anonymize=False, snr='auto'):
		"""
		Estimate the highest performance metrics achievable when predicting the :code:`target_column` using all other columns.

		When :code:`problem_type=None`, the nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`target_column` is categorical.


		Parameters
		----------
		target_column : str
			The name of the column containing true labels.
		problem_type : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.
		anonymize : bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost).



		Returns
		-------
		achievable_performance : pandas.Dataframe
			The result is a pandas.Dataframe with columns (where applicable):

			* :code:`'Achievable Accuracy'`: The highest classification accuracy that can be achieved by a model using provided inputs to predict the label.
			* :code:`'Achievable R^2'`: The highest :math:`R^2` that can be achieved by a model using provided inputs to predict the label.
			* :code:`'Achievable RMSE'`: The lowest Root Mean Square Error that can be achieved by a model using provided inputs to predict the label.		
			* :code:`'Achievable Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a model using provided inputs to predict the label.



		.. admonition:: Theoretical Foundation

			Section :ref:`1 - Achievable Performance`.


		.. seealso::

			:ref:`kxy.pre_learning.achievable_performance.data_valuation <data-valuation>`
			
		"""
		assert target_column in self._obj.columns, 'The target_column should be a column'
		if problem_type is None:
			problem_type = 'classification' if self.is_discrete(target_column) else 'regression'

		_obj = self.anonymize(columns_to_exclude=[target_column]) if anonymize or self.is_too_large else self._obj

		return dv(_obj, target_column, problem_type, snr=snr)


	def variable_selection(self, target_column, problem_type=None, anonymize=False, snr='auto'):
		"""
		Runs the model-free variable selection analysis.

		When :code:`problem_type=None`, the nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`target_column` is categorical.


		Parameters
		----------
		target_column : str
			The name of the column containing true labels.
		problem_type : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.
		anonymize : bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost).

		Returns
		-------
		result : pandas.DataFrame
			The result is a pandas.DataFrame with columns (where applicable):

			* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
			* :code:`'Variable'`: The column name corresponding to the input variable.
			* :code:`'Running Achievable R^2'`: The highest :math:`R^2` that can be achieved by a classification model using all variables selected so far, including this one.
			* :code:`'Running Achievable Accuracy'`: The highest classification accuracy that can be achieved by a classification model using all variables selected so far, including this one.
			* :code:`'Running Achievable RMSE'`: The highest classification accuracy that can be achieved by a classification model using all variables selected so far, including this one.


		.. admonition:: Theoretical Foundation

			Section :ref:`2 - Variable Selection Analysis`.

		.. seealso::

			:ref:`kxy.pre_learning.variable_selection.variable_selection <variable-selection>`
		"""
		assert target_column in self._obj.columns, 'The target_column should be a column'
		if problem_type is None:
			problem_type = 'classification' if self.is_discrete(target_column) else 'regression'

		_obj = self.anonymize(columns_to_exclude=[target_column]) if anonymize or self.is_too_large else self._obj

		return vs(_obj, target_column, problem_type, snr=snr)








#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pandas as pd

from kxy.api import upload_data
from kxy.post_learning import data_driven_improvability as ddi
from kxy.post_learning import model_driven_improvability as mdi
from kxy.post_learning import model_explanation as me

from .base_accessor import BaseAccessor

@pd.api.extensions.register_dataframe_accessor("kxy_post_learning")
class PostLearningAccessor(BaseAccessor):
	"""
	Extension of the pandas.DataFrame class with various analytics for **post-learning** in supervised learning problems.

	This class defines the :code:`kxy_post_learning` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_.

	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_post_learning.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
	"""

	def data_driven_improvability(self, target_column, new_variables, problem_type=None, anonymize=False, snr='auto'):
		"""
		Estimate the potential performance boost that a set of new explanatory variables can bring about.


		Parameters
		----------
		target_column : str
			The name of the column containing true labels.
		new_variables : list
			The names of the columns to use as new explanatory variables.
		problem_type : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from whether or not :code:`target_column` is categorical.
		anonymize : bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost).



		Returns
		-------
		result : pandas.Dataframe
			The result is a pandas.Dataframe with columns (where applicable):

			* :code:`'Accuracy Boost'`: The classification accuracy boost that the new explanatory variables can bring about.
			* :code:`'R-Squared Boost'`: The :math:`R^2` boost that the new explanatory variables can bring about.
			* :code:`'RMSE Reduction'`: The reduction in Root Mean Square Error that the new explanatory variables can bring about.
			* :code:`'Log-Likelihood Per Sample Boost'`: The boost in log-likelihood per sample that the new explanatory variables can bring about.


		.. admonition:: Theoretical Foundation

			Section :ref:`3 - Model Improvability`.


		.. seealso::

			:ref:`kxy.post_learning.improvability.data_driven_improvability <data-driven-improvability>`

		"""
		assert target_column in self._obj.columns, 'The target_column should be a column'
		if problem_type is None:
			problem_type = 'classification' if self.is_discrete(target_column) else 'regression'

		_obj = self.anonymize(columns_to_exclude=[target_column]) if anonymize or self.is_too_large else self._obj

		return ddi(_obj, target_column, new_variables, problem_type, snr=snr)


	def model_driven_improvability(self, target_column, prediction_column, problem_type=None, anonymize=False, snr='auto'):
		"""
		Estimate the extent to which a trained supervised learner may be improved in a model-driven fashion (i.e. without resorting to additional explanatory variables).


		Parameters
		----------
		target_column : str
			The name of the column containing true labels.
		prediction_column : str
			The name of the column containing model predictions.
		problem_type : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from whether or not :code:`target_column` is categorical.
		anonymize : bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost).


		Returns
		-------
		result : pandas.Dataframe
			The result is a pandas.Dataframe with columns (where applicable):

			* :code:`'Lost Accuracy'`: The amount of classification accuracy that was irreversibly lost when training the supervised learner.
			* :code:`'Lost R-Squared'`: The amount of :math:`R^2` that was irreversibly lost when training the supervised learner.
			* :code:`'Lost RMSE'`: The amount of Root Mean Square Error that was irreversibly lost when training the supervised learner.		
			* :code:`'Lost Log-Likelihood Per Sample'`: The amount of true log-likelihood per sample that was irreversibly lost when training the supervised learner.

			* :code:`'Residual R-Squared'`: For regression problems, this is the highest :math:`R^2` that may be achieved when using explanatory variables to predict regression residuals.
			* :code:`'Residual RMSE'`: For regression problems, this is the lowest Root Mean Square Error that may be achieved when using explanatory variables to predict regression residuals.
			* :code:`'Residual Log-Likelihood Per Sample'`: For regression problems, this is the highest log-likelihood per sample that may be achieved when using explanatory variables to predict regression residuals.


		.. admonition:: Theoretical Foundation

			Section :ref:`3 - Model Improvability`.


		.. seealso::

			:ref:`kxy.post_learning.improvability.model_driven_improvability <model-driven-improvability>`

		"""
		assert target_column in self._obj.columns, 'The target_column should be a column'
		assert prediction_column in self._obj.columns, 'The prediction_column should be a column'
		if problem_type is None:
			problem_type = 'classification' if self.is_discrete(target_column) else 'regression'

		_obj = self.anonymize(columns_to_exclude=[target_column, prediction_column]) if anonymize or self.is_too_large else self._obj

		return mdi(_obj, target_column, prediction_column, problem_type, snr=snr)


	def model_explanation(self, prediction_column, problem_type=None, anonymize=False, snr='auto'):
		"""
		Analyzes the variables that a model relies on the most in a brute-force fashion.
		
		The first variable is the variable the model relies on the most. The second variable is the variable that complements the first variable the most in explaining model decisions etc.

		Running performances should be understood as the performance achievable when trying to guess model predictions using variables with selection order smaller or equal to that of the row.

		When :code:`problem_type=None`, the nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`target_column` is categorical.


		Parameters
		----------
		prediction_column : str
			The name of the column containing model predictions.
		problem_type : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.
		anonymize : bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost).

		Returns
		-------
		result : pandas.DataFrame
			The result is a pandas.Dataframe with columns (where applicable):

			* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
			* :code:`'Variable'`: The column name corresponding to the input variable.
			* :code:`'Running Achievable R^2'`: The highest :math:`R^2` that can be achieved by a classification model using all variables selected so far, including this one.
			* :code:`'Running Achievable Accuracy'`: The highest classification accuracy that can be achieved by a classification model using all variables selected so far, including this one.
			* :code:`'Running Achievable RMSE'`: The highest classification accuracy that can be achieved by a classification model using all variables selected so far, including this one.


		.. admonition:: Theoretical Foundation

			Section :ref:`2 - Variable Selection Analysis`.


		.. seealso::

			:ref:`kxy.post_learning.model_explanation.model_explanation <variable-selection>`
			
		"""
		assert prediction_column in self._obj.columns, 'The prediction_column should be a column'
		if problem_type is None:
			problem_type = 'classification' if self.is_discrete(prediction_column) else 'regression'

		_obj = self.anonymize(columns_to_exclude=[prediction_column]) if anonymize or self.is_too_large else self._obj

		return me(_obj, prediction_column, problem_type, snr=snr)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
We define a custom :code:`kxy` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_ below, 
namely the class :code:`Accessor`, that extends the pandas DataFrame class with all our analyses, thereby allowing data scientists to tap into 
the power of the :code:`kxy` toolkit within the comfort of their favorite data structure.

All methods defined in the :code:`Accessor` class are accessible from any DataFrame instance as :code:`df.kxy.<method_name>`, so long as the :code:`kxy` python 
package is imported alongside :code:`pandas`. 
"""


import pandas as pd

from .features_accessor import FeaturesAccessor
from .finance_accessor import FinanceAccessor
from .learning_accessor import LearningAccessor
from .post_learning_accessor import PostLearningAccessor
from .pre_learning_accessor import PreLearningAccessor


@pd.api.extensions.register_dataframe_accessor("kxy")
class Accessor(PreLearningAccessor, LearningAccessor, PostLearningAccessor, FinanceAccessor, FeaturesAccessor):
	"""
	Extension of the pandas.DataFrame class with the full capabilities of the :code:`kxy` platform.
	"""
	pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-



def approx_opt_remaining_time(step):
	"""
	Estimate the amount of time left for the API to finish a task that requires maximum-entropy optimization.


	Parameters
	----------
	step: int
		A string that uniquely identifies the content of the file.

	Returns
	-------
	remaining_time : str
		The approximate remaining duration.
	"""
	elapsed_time = 10*step
	total_time = 10 if step < 1 else 2*60 if step <= 10 else 60*5 if step <= 25 else 60*10 if step <= 50 else 60*20 if step <= 80 else 60*30
	remaining_time = max(total_time-elapsed_time, 0)

	if step == 100:
		remaining_time = 0

	remaining_time = str(remaining_time) + 's   ' if remaining_time < 60 else str(remaining_time//60) + 'min '

	return remaining_time


def approx_beta_remaining_time(step):
	"""
	Estimate the amount of time left for the API to complete an information-adjusted correlation request.


	Parameters
	----------
	step: int
		A string that uniquely identifies the content of the file.

	Returns
	-------
	remaining_time : str
		The approximate remaining duration.
	"""
	elapsed_time = step
	total_time = 120
	remaining_time = max(total_time-elapsed_time, 0)

	if step == 100:
		remaining_time = 0

	remaining_time = str(remaining_time) + 's   ' if remaining_time < 60 else str(remaining_time//60) + 'min '

	return remaining_time





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Everything billing.
"""
import logging
logging.basicConfig(level=logging.INFO)
import requests
from time import time

from kxy.api import APIClient


def get_upcoming_invoice():
	"""
	Retrieves all items that will show up in your next invoice.

	Returns
	-------
	d : dict
		The dictionary containing all items that will appear in your next invoice.
		E.g. :code:`{'Type of charge': {'total_usd': ..., 'quantity': ..., 'description': ..., 'billing_period_start_timestamp': ..., 'billing_period_end_timestamp': ...}, ... }`
	"""
	api_response = APIClient.route(
			path='/wk/billing/upcoming-invoice', method='POST',\
			timestamp=int(time()))
	try:
		if api_response.status_code == requests.codes.ok:
			api_response = api_response.json()
			return api_response
		else:
			return {}
	except:
		logging.exception('Failed to retrieve your upcoming invoice.')
		return {}
